""" Helper functions for the `run` command. """

import importlib
import inspect

import dask
import edo


def get_default_optimiser_arguments():
    """ Get the default arguments from `edo.DataOptimiser`. """

    signature = inspect.signature(edo.DataOptimiser)
    defaults = {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }

    defaults["root"] = None
    defaults["processes"] = None
    defaults["fitness_kwargs"] = None
    defaults["stop_kwargs"] = None
    defaults["dwindle_kwargs"] = None

    return defaults


def get_experiment_parameters(experiment):
    """ Get the parameters for the experiment. """

    spec = importlib.util.spec_from_file_location("__main__", experiment)
    experiment = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(experiment)

    module = {k.lower(): v for k, v in vars(experiment).items()}

    all_params = set(inspect.getfullargspec(edo.DataOptimiser).args) | set(
        inspect.getfullargspec(edo.DataOptimiser.run).args
    )

    module_params = {k: v for k, v in module.items() if k in all_params}

    module_params["families"] = [
        edo.Family(dist) for dist in module["distributions"]
    ]
    module_params["optimiser"] = module.get("optimiser", edo.DataOptimiser)

    params = get_default_optimiser_arguments()
    params.update(module_params)

    return params


@dask.delayed
def run_single_trial(experiment, root, seed):
    """ Lazily run a single trial of an experiment. """

    params = get_experiment_parameters(experiment)

    _ = params.pop("root")
    optimiser = params.pop("optimiser")
    processes = params.pop("processes")
    fitness_kwargs = params.pop("fitness_kwargs")
    stop_kwargs = params.pop("stop_kwargs")
    dwindle_kwargs = params.pop("dwindle_kwargs")

    opt = optimiser(**params)
    _ = opt.run(
        root=root / str(seed),
        processes=processes,
        random_state=seed,
        fitness_kwargs=fitness_kwargs,
        stop_kwargs=stop_kwargs,
        dwindle_kwargs=dwindle_kwargs,
    )



""" The main script. """

import os
import pathlib
import tarfile

import click
import dask
import pandas as pd
import tqdm
from dask.diagnostics import ProgressBar

from .run import get_experiment_parameters, run_single_trial
from .summarise import (
    get_distributions,
    get_representative_idxs,
    get_trial_summary,
    write_representatives,
)
from .version import __version__


def _get_root_from_experiment(experiment):
    """Get the root directory from an experiment. If there isn't one, set the
    root to be adjacent to the experiment script."""

    root = get_experiment_parameters(experiment).pop("root")
    root = experiment.parent if root is None else root

    return root


@click.group(invoke_without_command=True)
@click.option(
    "--version", is_flag=True, default=False, help="The current version."
)
def main(version):
    """ Run and summarise experiments with `edo`. """

    if version:
        click.echo(__version__)


@main.command()
@click.argument("experiment", type=click.Path(exists=True))
@click.option(
    "--cores", default=None, type=int, help="The number of cores to use."
)
@click.option("--seeds", default=1, help="The number of trials to run.")
def run(experiment, cores, seeds):
    """ Run a series of trials using the `experiment` script. """

    experiment = pathlib.Path(experiment).resolve()
    name = experiment.stem

    click.echo(f"Running experiment: {name}")

    root = _get_root_from_experiment(experiment)
    out = pathlib.Path(root) / name / "data"
    out.mkdir(exist_ok=True, parents=True)

    click.echo(f"Writing to: {out}")

    tasks = (run_single_trial(experiment, out, seed) for seed in range(seeds))

    with ProgressBar():
        if cores is None:
            dask.compute(*tasks, scheduler="single-threaded")
        else:
            dask.compute(*tasks, num_workers=cores, scheduler="processes")

    click.echo("Experiment complete")


@main.command()
@click.option(
    "--tarball/--no-tarball",
    default=False,
    help="Tarball the data and delete original, or don't.",
)
@click.argument("experiment", type=click.Path(exists=True))
@click.argument("quantiles", nargs=-1, type=float, required=False)
def summarise(tarball, experiment, quantiles):
    """Summarise the EDO data from an experiment.

    Here, `experiment` should be a path to an experiment script of the form
    `/path/to/experiment/<experiment-name>.py`.

    To specify quantiles (between 0 and 1), list them at the end separated by
    spaces. Defaults to the minimum, median and maximum."""

    if not quantiles:
        quantiles = (0, 0.5, 1)

    experiment = pathlib.Path(experiment)
    root = _get_root_from_experiment(experiment)
    out = pathlib.Path(root) / experiment.stem
    data = out / "data"
    summary_path = out / "summary"
    summary_path.mkdir(exist_ok=True)

    click.echo(f"Summarising data at {out}")

    distributions = get_distributions(experiment)
    trials = [
        path
        for path in data.iterdir()
        if path.is_dir() and path.stem != "subtypes"
    ]
    summaries = []
    for trial in tqdm.tqdm(trials):
        summaries.append(get_trial_summary(trial, distributions))

    summary = pd.concat(summaries)
    summary = summary.sort_values(["generation", "individual", "seed"])
    summary.to_csv(summary_path / "main.csv", index=False)

    click.echo("Getting representative individuals")

    idxs = get_representative_idxs(summary, quantiles)
    write_representatives(summary, idxs, data, summary_path)

    if tarball:
        click.echo("Making tarball")

        with tarfile.open(str(data) + ".tar.gz", "w:gz") as tar:
            tar.add(data, arcname=data.stem)

        os.system(f"rm -r {data}")

    click.echo("Summary complete")


if __name__ == "__main__":  # pragma: no cover
    main()



""" Helper functions for the `summarise` command. """

import importlib
import os

import edo
import numpy as np
import pandas as pd


def get_distributions(experiment):
    """ Get the distribution classes used in the experiment. """

    spec = importlib.util.spec_from_file_location("__main__", experiment)
    experiment = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(experiment)

    return experiment.distributions


def get_trial_summary(trial, distributions):
    """Summarise a trial by recording the origins, shapes, sizes and fitnesses
    of the individuals created in that trial."""

    fitness = pd.read_csv(trial / "fitness.csv")
    max_generation = fitness["generation"].max()

    pop_history = edo.optimiser._get_pop_history(
        trial, max_generation + 1, distributions
    )

    summary_dfs = []
    for gen, generation in enumerate(pop_history):
        generation_summary = []
        for idx, individual in enumerate(generation):

            dataframe = individual.dataframe
            nrows = len(dataframe)
            ncols = len(dataframe.columns)
            size = dataframe.memory_usage().sum().compute()

            generation_summary.append((idx, nrows, ncols, size))

        generation_summary = pd.DataFrame(
            generation_summary,
            columns=["individual", "nrows", "ncols", "memory"],
        )
        generation_summary["generation"] = gen
        summary_dfs.append(generation_summary)

    summary = pd.concat(summary_dfs, axis=0, ignore_index=True)
    summary["fitness"] = fitness["fitness"]
    summary["seed"] = int(trial.stem)

    return summary


def get_representative_idxs(summary, quantiles):
    """Get the indices of those individuals that have the closest fitness to
    the given quantiles."""

    fitness = summary["fitness"]
    idxs = {}
    for quantile in quantiles:
        diffs = (fitness - fitness.quantile(quantile)).abs().values
        idxs[quantile] = np.argmin(diffs)

    return idxs


def write_representatives(summary, idxs, data_path, summary_path):
    """ Write to file the individuals at the indices given by `idxs`. """

    for quantile, idx in idxs.items():

        gen, ind, seed = map(
            str, summary[["generation", "individual", "seed"]].iloc[idx, :]
        )

        origin = data_path / seed / gen / ind
        out = summary_path / str(quantile)
        out.mkdir(exist_ok=True, parents=True)

        os.system(f"cp -r {origin}/main.* {out}")
        with open(out / "README", "w") as readme:
            readme.write(
                f"Invididual {ind} of generation {gen} in trial {seed}."
            )



""" Setup file. """

from setuptools import find_packages, setup

with open("README.md", "r") as readme:
    README = readme.read()

requirements = []
with open("requirements.txt", "r") as reqs:
    for req in reqs.read().splitlines():
        requirements.append(req)

version = {}
with open("src/edolab/version.py", "r") as vers:
    exec(vers.read(), version)


setup(
    name="edolab",
    version=version["__version__"],
    description="A command line tool for running experiments with `edo`.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/daffidwilde/edolab",
    author="Henry Wilde",
    author_email="henrydavidwilde@gmail.com",
    license="MIT",
    keywords=["artificial data", "evolutionary algorithm", "experiments"],
    packages=find_packages("src"),
    package_dir={"": "src"},
    python_requires=">=3.6",
    install_requires=requirements,
    tests_require=["pytest", "pytest-cov", "numpy"],
    entry_points={"console_scripts": "edolab=edolab.__main__:main"},
)
